#include "distrib.h"
#include "naivekd.h"
#include "optkd.h"
#include "sproullkd.h"
#include "vptree.h"
#include "nn.h"
#include "show.h"

#include <stdio.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <Xm/Xm.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>
#include <Xm/RowColumn.h>
#include <Xm/PushBG.h>
#include <Xm/Form.h>
#include <Xm/CascadeB.h>
#include <Xm/SelectioB.h>
#include <Xm/PushB.h>
#include <Xm/BulletinB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/MessageB.h>
#include <Xm/Label.h>
#include <Xm/FileSB.h>
#include <Xm/MainW.h>

#define MAXDIM 25

#include <Xm/Label.h>
#include <Xm/FileSB.h>
#include <Xm/MainW.h>

#define MAXDIM 25

int numpoints,
    dimension = MAXDIM;

int XProjDim = 0,
    YProjDim = 1,
    ZProjDim = 2;

float **Points;

TreePtr NaivekdRoot;  /* pointer to a naive kdtree */
optkdNode *OptkdRoot;
VPTreeNode *VPRoot;
sproullkdNode *SproullkdRoot;

enum datasource  { undefineddat,
                   file,
                   distribution }
                  CurrDataSource = undefineddat;

enum displaymode { undefineddisp, 
                   twod,
                   threed}
                  CurrDisplayMode = undefineddisp;

enum querymethod { undefinedquer,
                   rectangular,
                   nearestneighbor,
                   nearestneighborgraph}
                  CurrQueryMethod = undefinedquer;

enum datastruct { undefinedds,
                  naive,
                  naivekd,
                  optkd,
                  sproullkd,
                  vptree}
                 CurrDataStruct = undefinedds;

XGCValues gcv;
GC        gc;

/* from fileoper.c */
 extern float **ReadInputFile(); 
      
/* from distrib.c */
extern float **GenerateUniform();
extern float **GenerateAnnulus();
extern float **GenerateArith();
extern float **GenerateBall();
extern float **GenerateClusnorm(); 
extern float **GenerateCubediam();
extern float **GenerateCubeedge();
extern float **GenerateCorners();
extern float **GenerateGrid();
extern float **GenerateNormal();
extern float **GenerateSpokes();
extern float **GenerateHenon();

/* from queryparser.c */
extern float **ParseRectangular();
extern float *ParseNearestNeighbor();

/* from drawing.c */
    /* colors used by GL routines */
#define BLACK  0
#define RED    1
#define YELLOW 3 

extern void OpenGLWindow();
extern void DisplayPoints2d();
extern void DisplayPoints3d();
extern void DisplayRectQuery2d();
extern void DisplayRectQuery3d();
extern void DisplayNNQuery2d();
extern void DisplayNNQuery3d();
extern void ClearGLDisplay();
extern void UpdateGLDisplay();
extern void Draw2dNaivekd();
extern void Draw3dNaivekd();
extern void Draw2dOptkd();
extern void Draw3dOptkd();
extern void Draw2dSproullkd();
extern void DrawVPNode2d();

/* from naive.c */
extern int *NaiveRectQuery();
extern int *NaiveNNQuery();

/* from naivekd.c */
extern TreePtr kdNaiveBuildTree();
extern int *kdNaiveRectQuery();
extern void KillNaiveTree();

/* from optkd.c */
extern optkdNode *BuildOptTree();
extern int *kdOptNNQuery();
extern int *DrawkdOptNNQuery2d();
extern int *DrawkdOptNNQuery3d();
extern void KillOptTree();
extern int *kdOptRectQuery();

/* from sproullkd.c */
extern sproullkdNode *BuildSproullkdTree();
extern int *kdSproullNNQuery();
extern void KillSproullTree();

/* from vptree.c */
extern VPTreeNode *BuildVPTree();
extern int *vpNNQuery();
extern void KillVPtree();
float **(*DistribFun)();
            
/* from XWidgetsCreator */

extern Widget PullMeX,PullMeY,PullMeZ;
extern Widget FileBoard;
extern Widget DistribBoard;
extern Widget DisplayBoard;
extern Widget RectQBoard;
extern Widget NNQBoard;
extern Widget NNGBoard;
extern Widget fselbox;
extern Widget error;
extern Widget menubutt2, menubutt3, menubutt4, menubutt5;
extern Widget GenDatButt, DisplayDatButt, BuildDSButt, DisplayQuerButt,
              ClearDispButt;
extern Widget DimensionW, NumPointsW;
extern Widget XProjectCascade, YProjectCascade,ZProject,ZProjectCascade;
extern Widget DummyXProjButt;
extern Widget DummyYProjButt;
extern Widget RectangularButt,NNButt;
extern Widget RectQText;
extern Widget NNQuerPointText,NumPointsText;
extern Widget NNGNumPointsText;
extern Widget FDimText,FNumPText,FFileNText;
extern Widget RotateButts;
extern Widget MinkowP;

/* from Fortune's VorMain.c */
extern void DoVoronoi();

/* from graphs.c */
extern void DrawNNG2d();
extern void DrawNNG3d();

void KillExistingDataStructure();

Widget CurrDataWidget;
Widget CurrDisplayWidget;
Widget CurrQueryMethodWidget;

/* dynamic widgets */
Widget *DimButtsX;  
Widget *DimButtsY;
Widget *DimButtsZ;

XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

int ShowWhat = POINTS;
int Metric = EUCLIDEAN;
int Timing = 0;
/****************************************************************************/

void SetXProjDim(w, client_data, call_data)

/****************************************************************************/

/* Procedure Name   : SetXProjDim
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : error 
 * Description      : This callback sets the coordinate of the X projection.
 */

Widget	w;
int client_data;
caddr_t call_data;
{
  XProjDim =  client_data - 1;
}

/****************************************************************************/

void SetYProjDim(w, client_data, call_data)

/****************************************************************************/

/* Procedure Name   : SetXProjDim
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : error 
 * Description      : This callback sets the coordinate of the Y projection.
 */

Widget	w;
int client_data;
caddr_t call_data;
{
  YProjDim =  client_data - 1;
}

/****************************************************************************/

void SetZProjDim(w, client_data, call_data)

/****************************************************************************/

/* Procedure Name   : SetXProjDim
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : error 
 * Description      : This callback sets the coordinate of the Z projection.
 */

Widget	w;
int client_data;
caddr_t call_data;
{
  ZProjDim =  client_data - 1;
}

/***********************************************************************/

void InitialState()

/************************************************************************/

/* Procedure Name   : InitialState()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This initializes the ranger program.
 */

{
  int k,n;
  char s[10];     
  Arg args[8];

  XtUnmanageChild(FileBoard);
  XtUnmanageChild(DistribBoard);
  XtUnmanageChild(DisplayBoard);
  XtUnmanageChild(RectQBoard);
  XtUnmanageChild(NNQBoard);
  XtUnmanageChild(fselbox);
  XtUnmanageChild(error);
  XtSetSensitive(menubutt2,FALSE);
  XtSetSensitive(menubutt4,FALSE);
  XtSetSensitive(GenDatButt, FALSE);
  XtSetSensitive(DisplayDatButt, FALSE);
  XtSetSensitive(BuildDSButt, FALSE);
  XtSetSensitive(DisplayQuerButt, FALSE);
  XtSetSensitive(ClearDispButt,FALSE);
  DistribFun = GenerateUniform;  

  /* set up projection dimension widgets */
  DimButtsX = (Widget *) malloc(sizeof(Widget) * MAXDIM);
  DimButtsY = (Widget *) malloc(sizeof(Widget) * MAXDIM);
  DimButtsZ = (Widget *) malloc(sizeof(Widget) * MAXDIM);
  for (k=1; k <= MAXDIM; k++) {
    sprintf(s,"%d\0",k);
    n=0;
    XtSetArg(args[n],XmNlabelString, XmStringCreate(s, XmSTRING_DEFAULT_CHARSET)); n++;
    DimButtsX[k-1] = XmCreatePushButton(XProjectCascade, s, args,n);
    DimButtsY[k-1] = XmCreatePushButton(YProjectCascade, s, args,n);
    DimButtsZ[k-1] = XmCreatePushButton(ZProjectCascade, s, args,n);
    XtAddCallback(DimButtsX[k-1],XmNactivateCallback,SetXProjDim,k);
    XtAddCallback(DimButtsY[k-1],XmNactivateCallback,SetYProjDim,k); 
    XtAddCallback(DimButtsZ[k-1],XmNactivateCallback,SetZProjDim,k); 
    XtManageChild(DimButtsX[k-1]);
    XtManageChild(DimButtsY[k-1]);
    XtManageChild(DimButtsZ[k-1]);
  }  

  n=0;
  XtSetArg(args[n],XmNlabelString,XmStringCreateLtoR("1", charset));n++;
  XtSetValues(PullMeX,args,n);
  n=0;
  XtSetArg(args[n],XmNlabelString,XmStringCreateLtoR("2", charset));n++;
  XtSetValues(PullMeY,args,n);
  n=0;
  XtSetArg(args[n],XmNlabelString,XmStringCreateLtoR("3", charset));n++;
  XtSetValues(PullMeZ,args,n);
 
  OpenGLWindow();
  NaivekdRoot=NULL;
  OptkdRoot = NULL;
}


/**************************************************************************/

double current_time()

/*************************************************************************/
{
/* returns user time in microseconds */

	struct rusage rusage;
	register double usr_time, sys_time;
	
	(void)getrusage(RUSAGE_SELF, &rusage);
	/* calculate time in microseconds */
	usr_time = (double)rusage.ru_utime.tv_sec * 1.0E6
			+ (double)rusage.ru_utime.tv_usec;
	return(usr_time);
}


/*****************************************************************************/

void UpdateDynamicWidgets()

/*****************************************************************************/

{
  int k,n;
  char from_s[256];
  XrmValue from_value,to_value;
  static int OldDim2d = MAXDIM;

  if (OldDim2d >dimension) {
    for (k=dimension+1; k <=MAXDIM; k++) {
      XtUnmanageChild(DimButtsX[k-1]);
      XtUnmanageChild(DimButtsY[k-1]);
      XtUnmanageChild(DimButtsZ[k-1]);
    }
  }
  else {
    for (k=OldDim2d+1; k <=dimension; k++) {
      XtManageChild(DimButtsX[k-1]);
      XtManageChild(DimButtsY[k-1]);
      XtManageChild(DimButtsZ[k-1]);
    }
  }

  if (CurrDisplayMode==threed && dimension < 3) {
    XtUnmanageChild(ZProject);
    XtUnmanageChild(RotateButts);
    CurrDisplayMode=twod;
  }

  OldDim2d = dimension;
}

/*****************************************************************************/

void DisplayError(error_message)

/*****************************************************************************/

/* Procedure Name   : DisplayError()
 * Return Value     : nothing
 * Input Variables  : char *error_message  Routine specified error message
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This routine displays the specified error message.
 */

char	*error_message;
{
	Arg	args[1];
	XtSetArg(args[0], XmNmessageString, XmStringCreateLtoR(error_message, charset));
	XtSetValues(error, args, 1);
	XtManageChild(error);
}

/****************************************************************************/

void ClearErrorMessage(w, client_data, call_data)

/****************************************************************************/

/* Procedure Name   : ClearErrorMessage()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : error 
 * Description      : This callback removes an error message.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;
{
    XtUnmanageChild(error);
}

/***********************************************************************/

void ExitRanger(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : ExitRanger()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback exits the ranger program.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;
{
      exit(0);
}

/***********************************************************************/

void RemoveFSelbox(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : RemoveFSelbox()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback removes the fileselectionbox.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  XtUnmanageChild(fselbox);
}

/***********************************************************************/

void ReadDataFile(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : ReadDataFile()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the FileBoard widget on the
 *                    ControlBoard, deactivating any other Data widget
 *                    currently selected.
 */

{
  Arg	args[1];
  int access_mode, access_ret,n;
  XmFileSelectionBoxCallbackStruct *fscb;
  char *filename, fnstr[140],dimlabel[50],numplabel[40];

  fscb = (XmFileSelectionBoxCallbackStruct *) call_data;
  XmStringGetLtoR(fscb->value,charset,&filename);
  access_mode = R_OK;
  access_ret=access(filename,access_mode);

  if (access_ret !=0) {
    XtUnmanageChild(fselbox);
    DisplayError("Cannot open file"); 
    return;
  }
  XtUnmanageChild(fselbox);
  Points=ReadInputFile(filename,&dimension, &numpoints); 

  if (dimension < 0) {
    XtUnmanageChild(fselbox);
    DisplayError("Cannot open file"); 
    return;
  }

  UpdateDynamicWidgets();
   
  if (Timing) {
    printf("Data read from external file named %s, with dimension %d, and %d points.\n",filename,dimension,numpoints);
  }

  if (CurrDataSource != undefineddat) {
    XtUnmanageChild(CurrDataWidget);
    XtSetSensitive(GenDatButt, FALSE);
  }

  CurrDataWidget=FileBoard;

  n=0;
  filename= strrchr(filename,'/');
  filename++;
  sprintf(fnstr,"%s","Filename: ");
  strcat(fnstr,filename);
  XtSetArg(args[n],XmNlabelString,XmStringCreateLtoR(fnstr, charset));n++;
  XtSetValues(FFileNText,args,n);

  n=0;
  sprintf(dimlabel,"%s %d","Dimension: ",dimension);
  XtSetArg(args[n],XmNlabelString,XmStringCreateLtoR(dimlabel,charset));n++;
  XtSetValues(FDimText,args,n);

  n=0;
  sprintf(numplabel,"%s %d","# of Points: ",numpoints);
  XtSetArg(args[n],XmNlabelString,XmStringCreateLtoR(numplabel,charset));n++;
  XtSetValues(FNumPText,args,n);
  XtManageChild(FileBoard);
  CurrDataSource = file;
  XtSetSensitive(menubutt2, TRUE);
}

/***********************************************************************/

void DataFromFile(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : DataFromFile()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the FileBoard widget on the
                      ControlBoard, deactivating any other Data widget
                      currently selected.
 */
Widget	w;
caddr_t client_data;
caddr_t call_data;
{
  XtManageChild(fselbox);
}

/***********************************************************************/

void DataFromDistribution(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : DataFromDistribution()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the DistribBoard widget on the
                      ControlBoard, deactivating any other DataBoard widget
                      currently selected.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  if (CurrDataSource != undefineddat) {
    XtUnmanageChild(CurrDataWidget);
  }
  CurrDataWidget=DistribBoard;
  XtManageChild(DistribBoard);
  CurrDataSource = distribution;
  XtSetSensitive(GenDatButt, TRUE);
}


/***********************************************************************/

void ShowMode(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : Show()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : I really don't know.
 */

Widget	w;
int client_data;
caddr_t call_data;

{
ShowWhat = client_data;
}


/***********************************************************************/

void Display2d(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : Display2d()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the Display2d widget on the
                      ControlBoard, deactivating any other DisplayBoard widget
                      currently selected.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  if (CurrDisplayMode == twod) {
    return;
  }

  if (CurrDisplayMode != undefineddisp) {
    XtUnmanageChild(CurrDisplayWidget);
  }
  
  CurrDisplayMode = twod; 
  CurrDisplayWidget=DisplayBoard;
  XtUnmanageChild(ZProject);
  XtUnmanageChild(RotateButts);
  XtManageChild(CurrDisplayWidget);
  UpdateDynamicWidgets();
  XtSetSensitive(DisplayDatButt,TRUE);
  XtSetSensitive(ClearDispButt,TRUE);
}


/***********************************************************************/

void Display3d(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : Display3d()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the Display3d widget on the
                      ControlBoard, deactivating any other DisplayBoard widget
                      currently selected.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{

  if (dimension < 3) {
    DisplayError("Dimension must be at least 3 for 3d projection!");
    return;
  } 


  if (CurrDisplayMode != undefineddisp) {
    XtUnmanageChild(CurrDisplayWidget);
  }

  CurrDisplayMode = threed;
  CurrDisplayWidget=DisplayBoard;
  XtManageChild(CurrDisplayWidget);
  XtManageChild(ZProject);
  XtManageChild(RotateButts);
  UpdateDynamicWidgets(); 
  XtSetSensitive(DisplayDatButt);
  XtSetSensitive(ClearDispButt);
}

/***********************************************************************/

void QueryRectangular(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : QueryRectangular()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the RectQBoard widget on the
                      ControlBoard, deactivating any other QueryBoard widget
                      currently selected.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  if (CurrQueryMethod != undefinedquer) {
    XtUnmanageChild(CurrQueryMethodWidget);
  }
  CurrQueryMethodWidget=RectQBoard;
  XtManageChild(CurrQueryMethodWidget);
  CurrQueryMethod = rectangular;
  XtSetSensitive(BuildDSButt, TRUE);
}

/***********************************************************************/

void QueryNN(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : QueryNN()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the NNQBoard widget on the
                      ControlBoard, deactivating any other QueryBoard widget
                      currently selected.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  if (CurrQueryMethod != undefinedquer) {
    XtUnmanageChild(CurrQueryMethodWidget);
  }
  CurrQueryMethodWidget=NNQBoard;
  XtManageChild(CurrQueryMethodWidget);
  CurrQueryMethod = nearestneighbor;
  XtSetSensitive(BuildDSButt, TRUE);
}

/***********************************************************************/

void QueryNNGraph(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : QueryNNGraph()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback enables the NNGBoard widget on the
                      ControlBoard, deactivating any other QueryBoard widget
                      currently selected.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  if (CurrQueryMethod != undefinedquer) {
    XtUnmanageChild(CurrQueryMethodWidget);
  }
  CurrQueryMethodWidget=NNGBoard;
  XtManageChild(CurrQueryMethodWidget);
  CurrQueryMethod = nearestneighborgraph;
  XtSetSensitive(BuildDSButt, TRUE);
}

/**********************************************************************/

void DistanceFun(w, client_data, call_data)

/**********************************************************************/

Widget	w;
int client_data;
caddr_t call_data;

{
  Metric = client_data;
  if (Metric == L_P) {
    XtManageChild(MinkowP);
  }
  else {
    XtUnmanageChild(MinkowP);
  }

}


/***********************************************************************/

void GenerateDist(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : GetDimension
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This procedure reads the dimension and number of points
 *                    from the DistribBoard widget.
 */

Widget	w;
int client_data;
caddr_t call_data;

{
switch (client_data) {
  case UNIFORM  : if (Timing) {printf("Data from uniform distribution\n"); }
                  DistribFun = GenerateUniform;
                  break;
  case ANNULUS  : if (Timing) {printf("Data from annulus distribution\n"); }
                  DistribFun = GenerateAnnulus;       
                  break;
  case ARITH    : if (Timing) {printf("Data from arith distribution\n"); }
                  DistribFun = GenerateArith;
                  break;
  case BALL     : if (Timing) {printf("Data from ball distribution\n"); }
                  DistribFun = GenerateBall;
                  break;
  case CLUSNORM : if (Timing) {printf("Data from clusnorm distribution\n"); }
                  DistribFun = GenerateClusnorm;
                  break;
  case CUBEDIAM : if (Timing) {printf("Data from cubediam distribution\n"); }
                  DistribFun = GenerateCubediam;
                  break;
  case CUBEEDGE : if (Timing) {printf("Data from cubeedge distribution\n"); }
                  DistribFun = GenerateCubeedge;
                  break;
  case CORNERS  : if (Timing) {printf("Data from corners distribution\n"); }
                  DistribFun = GenerateCorners; 
                  break;
  case GRID     : if (Timing) {printf("Data from grid distribution\n"); }
                  DistribFun = GenerateGrid;
                  break;
  case NORMAL   : if (Timing) {printf("Data from normal distribution\n"); }
                  DistribFun = GenerateNormal;
                  break;
  case SPOKES   : if (Timing) {printf("Data from spokes distribution\n"); }
                  DistribFun = GenerateSpokes; 
                  break;
  case HENON    : if (Timing) {printf("Data from henon distribution\n"); }
                  DistribFun = GenerateHenon; 
                  break;
  default       : DisplayError("Invalid Distribution selected");
  }
}

/***********************************************************************/

void GetDistribBoardData()

/************************************************************************/

/* Procedure Name   : GetDimension
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This procedure reads the dimension and number of points
 *                    from the DistribBoard widget.
 */

{
  int dimtry;
  int stat;

  stat=sscanf(XmTextGetString(DimensionW),"%d",&dimtry);
  if (stat==1) {
    if (dimtry > MAXDIM) {
      DisplayError("Dimension is too large");
      return;
    }
    if (dimtry < 2) {
      DisplayError("Dimension is too small");
      return;
    }
  }
  else
    {
      DisplayError("Dimension invalid");
      return;
    }
  dimension=dimtry;

  stat=sscanf(XmTextGetString(NumPointsW),"%d",&numpoints);
  if (stat==1) {
    if (numpoints > 1000000) {
      DisplayError("Too many points");
      return;
    }
    if (numpoints < 1) {
      DisplayError("Too few points");
      return;
    }
  }
  else
    {
      DisplayError("Number of points invalid");
      return;
    }

}
/***********************************************************************/

void GenerateData(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : GenerateData()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the GenDatButton is
 *                    pressed.  It reads the dimension and number of points
 *                    and calls the appropriate function to generate the
 *                    data.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  GetDistribBoardData();
  
  free(Points);
  Points = (*DistribFun)(numpoints,dimension);
  /* insert code here to call the proper distribution function */
 
  XtSetSensitive(GenDatButt, TRUE);
  XtSetSensitive(menubutt2, TRUE);
  UpdateDynamicWidgets();
  if (Timing) {printf("dimension=%d numpoints=%d\n",dimension,numpoints);}

  KillExistingDataStructure();
}

/***********************************************************************/

void DisplayData(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : DisplayData()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the DispDatButton is
 *                    pressed.  It reads the XProj and YProj dimension and
 *                    number of points and calls the appropriate function
 *                    to display the data.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  if (XProjDim > dimension) {
    DisplayError("X Projection Dimension not set properly");
    return;
  }

  if (YProjDim > dimension) {
    DisplayError("Y Projection Dimension not set properly");
    return;
  }

  if (CurrDisplayMode ==threed && ZProjDim > dimension) {
    DisplayError("Z Projection Dimension not set properly");
    return;
  }

  if (XProjDim == YProjDim) {
    DisplayError("X and Y Projection Dimensions cannot be equal");
    return;
  }

  if (CurrDisplayMode == threed) {
    if (ZProjDim > dimension) {
      DisplayError("Z Projection Dimension not set properly");
      return;
    }
    if (XProjDim == ZProjDim) {
      DisplayError("X and Z Projection Dimensions cannot be equal");
      return; 
    }
    if (ZProjDim == YProjDim) {
      DisplayError("Y and Z Projection Dimensions cannot be equal");
      return;
    }
  }

  if (CurrDisplayMode == twod) {
    switch(ShowWhat) {
      case POINTS :
        DisplayPoints2d(Points,numpoints,XProjDim,YProjDim,YELLOW); 
        return;
      case DATASTRUCT:   
        switch(CurrDataStruct) {
          case naivekd:
            if (NaivekdRoot != NULL) {
              Draw2dNaivekd(NaivekdRoot,Points,XProjDim,YProjDim);
            } else {
              NaivekdRoot = kdNaiveBuildTree(Points,numpoints,dimension);
              ClearGLDisplay();
              DisplayPoints2d(Points,numpoints,XProjDim,YProjDim,YELLOW); 
              Draw2dNaivekd(NaivekdRoot,Points,XProjDim,YProjDim);
            }            
            break;
          case optkd:
             if (OptkdRoot != NULL) {
               Draw2dOptkd(OptkdRoot,Points,XProjDim,YProjDim);
             }  else {
              OptkdRoot = BuildOptTree(Points,numpoints,dimension);
              ClearGLDisplay();
              DisplayPoints2d(Points,numpoints,XProjDim,YProjDim,YELLOW); 
              Draw2dOptkd(OptkdRoot,Points,XProjDim,YProjDim);
/*              DisplayError("Can't show data structure at this time"); */
            }     
            break;       
          case sproullkd:
            if (SproullkdRoot != NULL) {
              Draw2dSproullkd(SproullkdRoot,Points,XProjDim,YProjDim);
            }  else {
              SproullkdRoot=BuildSproullkdTree(Points,numpoints,dimension);
              ClearGLDisplay();
              DisplayPoints2d(Points,numpoints,XProjDim,YProjDim,YELLOW); 
              Draw2dSproullkd(SproullkdRoot,Points,XProjDim,YProjDim);
/*              DisplayError("Can't show data structure at this time"); */
            }
            break;           
          case vptree:
            if (VPRoot != NULL) {
              DrawVPTree2d(VPRoot,Points);
            }  else {
              VPRoot=BuildVPTree(Points,numpoints,dimension);
              DisplayPoints2d(Points,numpoints,XProjDim,YProjDim,YELLOW); 
              DrawVPTree2d(VPRoot,Points);
/*              DisplayError("Can't show data structure at this time"); */
            }
            break;           
	  }
        break;
      case PROJECTION :
        ClearGLDisplay();
        DisplayProjectedPoints2d(Points,numpoints,dimension,YELLOW);
        UpdateGLDisplay();
        break;
      case DELAUNAY :
        ClearGLDisplay();
        DoVoronoi(numpoints,Points,XProjDim,YProjDim,1);
        UpdateGLDisplay();
        break;
    }
  }

  if (CurrDisplayMode == threed) {
    switch(ShowWhat) {
      case POINTS:
        DisplayPoints3d(Points,numpoints,XProjDim,YProjDim,ZProjDim,YELLOW); 
        return;
      case DATASTRUCT:
        switch(CurrDataStruct) {
          case naivekd:
            if (NaivekdRoot != NULL) {
              Draw3dNaivekd(NaivekdRoot,Points,XProjDim,YProjDim,ZProjDim);
            }
            else {
              DisplayError("Can't show data structure at this time");
            }
            break;
          case optkd:
            if (OptkdRoot != NULL) {          
              Draw3dOptkd(OptkdRoot,Points,XProjDim,YProjDim,ZProjDim);
            }
            else {
              DisplayError("Can't show data structure at this time");
            }
            break;
	  }
        break;
      default:
        DisplayError("Cannot show this structure in 3d");
      }
  }
}

/***********************************************************************/

void BuildDataStruct(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : BuildDataStruct()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the BuildDataStruct is
 *                    pressed.  It calls the appropriate function to build
 *                    the data structure. 
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  double t1,t2;

  if (CurrDataStruct == naivekd) {
    t1=current_time();
    NaivekdRoot = kdNaiveBuildTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("Naivekd preprocessing took %f\n",t2-t1);}
  }

  if (CurrDataStruct == optkd) {
    t1=current_time();
    OptkdRoot = BuildOptTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("Optkd preprocessing took %f\n",t2-t1);}
  }

  if (CurrDataStruct == sproullkd) {
    t1=current_time();
    SproullkdRoot=BuildSproullkdTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("Sproullkd preprocessing took %f\n",t2-t1);}
  }

  if (CurrDataStruct == vptree) {
    t1=current_time();
    VPRoot=BuildVPTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("vptree preprocessing took %f\n",t2-t1);}
  }

  XtSetSensitive(DisplayQuerButt, TRUE);
}

/***********************************************************************/

void DisplayQuery(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : DisplayQuery()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the Display Query Button
 *                    is pressed.  It reads the dimension and number of
 *                    points and calls the appropriate function to generate
 *                    the data.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{ char *rectstring,*nnstring;
  float **rectquery,*querpoint;
  int *Found,k,NumNN,MinkP;
  int *(*NNFunc)();
  void *root;
  double t1,t2;

  if (CurrQueryMethod == rectangular) {
    rectstring = XmTextFieldGetString(RectQText);
    rectquery = ParseRectangular(rectstring,dimension);
    if (rectquery == NULL) {
      DisplayError("Invalid rectangular query given");
      return;
    }
    switch(CurrDataStruct) {
      case naive:
        t1=current_time();        
        Found=NaiveRectQuery(Points,numpoints,dimension,rectquery); 
        t2=current_time();
        if (Timing) {printf("Naive rect query took %f\n",t2-t1);}
        break;

      case naivekd:
        t1=current_time();        
        Found = kdNaiveRectQuery(NaivekdRoot,Points,dimension,rectquery);
        t2=current_time();
        if (Timing) {printf("Naivekd rect query took %f\n",t2-t1);}
        break;

      case optkd:
        t1=current_time();        
        Found = kdOptRectQuery(OptkdRoot,Points,dimension,rectquery);
        t2=current_time();
        if (Timing) {printf("optkd rect query took %f\n",t2-t1);}
        break;
     }

    if (CurrDisplayMode == twod) {
      DisplayRectQuery2d(Points,Found,rectquery,XProjDim,YProjDim,RED);
    }

    if (CurrDisplayMode == threed) {
      DisplayRectQuery3d(Points,Found,rectquery,XProjDim,YProjDim,ZProjDim,RED);
    }  
  } 

  if (CurrQueryMethod == nearestneighbor) {
    /* get Minko metric */
    if (Metric == L_P) {
      if ((sscanf(XmTextFieldGetString(MinkowP),"%d",&MinkP) != 1) || MinkP < 1){
        DisplayError("P must be a positive integer");
        return;
      }
    }       

    if ((sscanf(XmTextFieldGetString(NumPointsText),"%d",&NumNN) != 1) || (NumNN <1) || 
        (NumNN > numpoints)) {
       DisplayError("Number of nnq points is invalid");
       return;
     }
    
     nnstring=XmTextFieldGetString(NNQuerPointText);
     
     querpoint=ParseNearestNeighbor(nnstring,dimension);
     if (querpoint == NULL) {
        DisplayError("Query point is incorrectly specified");
        return;
     }

     switch(CurrDataStruct) {
       case naive:
         t1=current_time();        
         Found = NaiveNNQuery(Points,numpoints,dimension,querpoint,NumNN,Metric,MinkP);
         t2=current_time();
         if (Timing) {printf("Naive NN query took %f\n",t2-t1);}
         break;
       case optkd:
         if (ShowWhat==DATASTRUCT) {
           if (CurrDisplayMode==twod) {
             Found = DrawkdOptNNQuery2d(Points,numpoints,dimension,querpoint,NumNN,Metric,OptkdRoot, MinkP,XProjDim,YProjDim);
           } else if (CurrDisplayMode==threed) {
             Found = DrawkdOptNNQuery3d(Points,numpoints,dimension,querpoint,NumNN,Metric,OptkdRoot, MinkP,XProjDim,YProjDim,ZProjDim);           
           }
         } else {

           t1=current_time();        
           Found = kdOptNNQuery(Points,dimension,querpoint,NumNN,Metric,OptkdRoot,MinkP);
           t2=current_time();
           if (Timing) {printf("Optkd NN query took %f\n",t2-t1);}
         }
         break;
       case sproullkd:
         t1=current_time();        
         Found = kdSproullNNQuery(Points,dimension,querpoint,NumNN,Metric,SproullkdRoot,MinkP);
         t2=current_time();
         if (Timing) {printf("Sproullkd NN query took %f\n",t2-t1);}
         break;
       case vptree:
         t1=current_time();        
         Found = vpNNQuery(Points,dimension,querpoint,NumNN,Metric,VPRoot);
         t2=current_time();
         if (Timing) {printf("vptree NN query took %f\n",t2-t1);}
         break;
         break;
     }

    if (CurrDisplayMode == twod) {
      DisplayNNQuery2d(Points,Found,querpoint,XProjDim,YProjDim,RED);
    }

    if (CurrDisplayMode == threed) {
      DisplayNNQuery3d(Points,Found,querpoint,XProjDim,YProjDim,ZProjDim,RED);
    }
  }

  if (CurrQueryMethod == nearestneighborgraph) {
    if ((sscanf(XmTextFieldGetString(NNGNumPointsText),"%d",&NumNN) != 1) || (NumNN <1) || 
        (NumNN > numpoints)) {
       DisplayError("Number of nnq points is invalid");
       return;
     }
    switch(CurrDataStruct) {
       case naive:
         break;
       case optkd:
         root=OptkdRoot;
         break;
       case sproullkd:
         root=SproullkdRoot;
         break;
       case vptree:
         root=VPRoot;
         break;
    }
    if (CurrDisplayMode == twod) {
      DrawNNG2d(Points,numpoints,dimension,NumNN,Metric,XProjDim,YProjDim,root,MinkP,CurrDataStruct);
    }
    if (CurrDisplayMode == threed) {
      DrawNNG3d(Points,numpoints,dimension,NumNN,Metric,XProjDim,YProjDim,ZProjDim,MinkP,root,CurrDataStruct);
    }

  }
  free(Found);

}

/*************************************************************************/

void KillExistingDataStructure()

/*************************************************************************/

{
  XtSetSensitive(DisplayQuerButt, FALSE);
  if (CurrDataStruct == naivekd) {
    KillNaiveTree(NaivekdRoot);
    NaivekdRoot=NULL;
  }
  if (CurrDataStruct == optkd) {
    KillOptTree(OptkdRoot);
    OptkdRoot=NULL;
  }
  if (CurrDataStruct == sproullkd) {
    KillSproullTree(SproullkdRoot);
    SproullkdRoot=NULL;
  }
  if (CurrDataStruct == vptree) {
    KillVPTree(VPRoot);
    VPRoot=NULL;
  }
}

/***********************************************************************/

void Naive(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : Naive
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the Naive Option is 
 *                    selected from the Algorithm menu bar.  It specifies
 *                    that the naive algorithm is the method of choice.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  KillExistingDataStructure();
  XtSetSensitive(menubutt4,TRUE);
  XtSetSensitive(RectangularButt,TRUE);
  XtSetSensitive(NNButt,TRUE);
  XtSetSensitive(BuildDSButt);
  CurrDataStruct = naive;
}

/***********************************************************************/

void NaivekdTree(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : kDTree
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the kdTree Option is 
 *                    selected from the Algorithm menu bar.  It specifies
 *                    that the kdTree algorithm is the method of choice.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{

  KillExistingDataStructure();
  if (CurrDataSource != undefineddat) {
    XtSetSensitive(menubutt4,TRUE);
  }

  XtSetSensitive(RectangularButt,TRUE);
  XtSetSensitive(NNButt,TRUE);
  XtSetSensitive(BuildDSButt);

  CurrDataStruct = naivekd; 
}

/***********************************************************************/

void OptKdTree(w, client_data, call_data)

/************************************************************************/

/* Procedure Name   : kDTree
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the kdTree Option is 
 *                    selected from the Algorithm menu bar.  It specifies
 *                    that the kdTree algorithm is the method of choice.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{

  KillExistingDataStructure();
  if (CurrDataSource != undefineddat) {
    XtSetSensitive(menubutt4,TRUE);
  }

  XtSetSensitive(RectangularButt,TRUE);
  XtSetSensitive(NNButt,TRUE);
  XtSetSensitive(BuildDSButt);

  CurrDataStruct = optkd; 
}

/***********************************************************************/

void Sproullkdtree(w, client_data, call_data)

/************************************************************************/

/* procedure Name   : kDTree
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the kdTree Option is 
 *                    selected from the Algorithm menu bar.  It specifies
 *                    that the kdTree algorithm is the method of choice.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  KillExistingDataStructure();
  if (CurrDataSource != undefineddat) {
    XtSetSensitive(menubutt4,TRUE);
  }

  XtSetSensitive(RectangularButt,TRUE);
  XtSetSensitive(NNButt,TRUE);
  XtSetSensitive(BuildDSButt);

  CurrDataStruct = sproullkd; 
}


/***********************************************************************/

void VPtree(w, client_data, call_data)

/************************************************************************/

/* procedure Name   : kDTree
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the kdTree Option is 
 *                    selected from the Algorithm menu bar.  It specifies
 *                    that the kdTree algorithm is the method of choice.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
  KillExistingDataStructure();
  if (CurrDataSource != undefineddat) {
    XtSetSensitive(menubutt4,TRUE);
  }

  XtSetSensitive(RectangularButt,TRUE);
  XtSetSensitive(NNButt,TRUE);
  XtSetSensitive(BuildDSButt);

  CurrDataStruct = vptree;
}


/***********************************************************************/

void TimingToggleCB(w, client_data, call_data)

/************************************************************************/

/* procedure Name   : kDTree
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the kdTree Option is 
 *                    selected from the Algorithm menu bar.  It specifies
 *                    that the kdTree algorithm is the method of choice.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{
if (!Timing) { /* timing should be turned on */
  Timing=1;
}
else {
  Timing=0;
}
}





