#include "distrib.h"
#include "naivekd.h"
#include "optkd.h"
#include "sproullkd.h"
#include "vptree.h"
#include "nn.h"
#include "show.h"

#include <stdio.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/resource.h>
int numpoints,
    dimension = MAXDIM;

float **Points;

TreePtr NaivekdRoot;  /* pointer to a naive kdtree */
optkdNode *OptkdRoot;
VPTreeNode *VPRoot;
sproullkdNode *SproullkdRoot;
      
/* from distrib.c */
extern float **GenerateUniform();
extern float **GenerateAnnulus();
extern float **GenerateArith();
extern float **GenerateBall();
extern float **GenerateClusnorm(); 
extern float **GenerateCubediam();
extern float **GenerateCubeedge();
extern float **GenerateCorners();
extern float **GenerateGrid();
extern float **GenerateNormal();
extern float **GenerateSpokes();
extern float **GenerateHenon();

/* from naive.c */
extern int *NaiveRectQuery();
extern int *NaiveNNQuery();

/* from naivekd.c */
extern TreePtr kdNaiveBuildTree();
extern int *kdNaiveRectQuery();
extern void KillNaiveTree();

/* from optkd.c */
extern optkdNode *BuildOptTree();
extern int *kdOptNNQuery();
extern int *DrawkdOptNNQuery2d();
extern int *DrawkdOptNNQuery3d();
extern void KillOptTree();
extern int *kdOptRectQuery();

/* from sproullkd.c */
extern sproullkdNode *BuildSproullkdTree();
extern int *kdSproullNNQuery();
extern void KillSproullTree();

/* from vptree.c */
extern VPTreeNode *BuildVPTree();
extern int *vpNNQuery();
extern void KillVPtree();
float **(*DistribFun)();

void KillExistingDataStructure();

int ShowWhat = POINTS;
int Metric = EUCLIDEAN;
int Timing = 0;

/**************************************************************************/

double current_time()

/*************************************************************************/
{
/* returns user time in microseconds */

	struct rusage rusage;
	register double usr_time, sys_time;
	
	(void)getrusage(RUSAGE_SELF, &rusage);
	/* calculate time in microseconds */
	usr_time = (double)rusage.ru_utime.tv_sec * 1.0E6
			+ (double)rusage.ru_utime.tv_usec;
	return(usr_time);
}


/***********************************************************************/

void BuildDataStruct()

/************************************************************************/

/* Procedure Name   : BuildDataStruct()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the BuildDataStruct is
 *                    pressed.  It calls the appropriate function to build
 *                    the data structure. 
 */

{
  double t1,t2;

  if (CurrDataStruct == naivekd) {
    t1=current_time();
    NaivekdRoot = kdNaiveBuildTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("Naivekd preprocessing took %f\n",t2-t1);}
  }

  if (CurrDataStruct == optkd) {
    t1=current_time();
    OptkdRoot = BuildOptTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("Optkd preprocessing took %f\n",t2-t1);}
  }

  if (CurrDataStruct == sproullkd) {
    t1=current_time();
    SproullkdRoot=BuildSproullkdTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("Sproullkd preprocessing took %f\n",t2-t1);}
  }

  if (CurrDataStruct == vptree) {
    t1=current_time();
    VPRoot=BuildVPTree(Points,numpoints,dimension);
    t2=current_time();
    if (Timing) {printf("vptree preprocessing took %f\n",t2-t1);}
  }

  XtSetSensitive(DisplayQuerButt, TRUE);
}

/***********************************************************************/

void DisplayQuery()

/************************************************************************/

/* Procedure Name   : DisplayQuery()
 * Return Value     : nothing
 * Input Variables  : Widget w            - Callback widget
 *                    caddr_t client_data - client data
 *                    caddr_t call_data   - call data
 * Output Variables : none
 * Update Variables : none
 * Global Variables : none
 * Description      : This callback is enabled when the Display Query Button
 *                    is pressed.  It reads the dimension and number of
 *                    points and calls the appropriate function to generate
 *                    the data.
 */

Widget	w;
caddr_t client_data;
caddr_t call_data;

{ char *rectstring,*nnstring;
  float **rectquery,*querpoint;
  int *Found,k,NumNN,MinkP;
  int *(*NNFunc)();
  void *root;
  double t1,t2;

  if (CurrQueryMethod == rectangular) {
    rectstring = XmTextFieldGetString(RectQText);
    rectquery = ParseRectangular(rectstring,dimension);
    if (rectquery == NULL) {
      DisplayError("Invalid rectangular query given");
      return;
    }
    switch(CurrDataStruct) {
      case naive:
        t1=current_time();        
        Found=NaiveRectQuery(Points,numpoints,dimension,rectquery); 
        t2=current_time();
        if (Timing) {printf("Naive rect query took %f\n",t2-t1);}
        break;

      case naivekd:
        t1=current_time();        
        Found = kdNaiveRectQuery(NaivekdRoot,Points,dimension,rectquery);
        t2=current_time();
        if (Timing) {printf("Naivekd rect query took %f\n",t2-t1);}
        break;

      case optkd:
        t1=current_time();        
        Found = kdOptRectQuery(OptkdRoot,Points,dimension,rectquery);
        t2=current_time();
        if (Timing) {printf("optkd rect query took %f\n",t2-t1);}
        break;
     }

    if (CurrDisplayMode == twod) {
      DisplayRectQuery2d(Points,Found,rectquery,XProjDim,YProjDim,RED);
    }

    if (CurrDisplayMode == threed) {
      DisplayRectQuery3d(Points,Found,rectquery,XProjDim,YProjDim,ZProjDim,RED);
    }  
  } 

  if (CurrQueryMethod == nearestneighbor) {
    /* get Minko metric */
    if (Metric == L_P) {
      if ((sscanf(XmTextFieldGetString(MinkowP),"%d",&MinkP) != 1) || MinkP < 1){
        DisplayError("P must be a positive integer");
        return;
      }
    }       

    if ((sscanf(XmTextFieldGetString(NumPointsText),"%d",&NumNN) != 1) || (NumNN <1) || 
        (NumNN > numpoints)) {
       DisplayError("Number of nnq points is invalid");
       return;
     }
    
     nnstring=XmTextFieldGetString(NNQuerPointText);
     
     querpoint=ParseNearestNeighbor(nnstring,dimension);
     if (querpoint == NULL) {
        DisplayError("Query point is incorrectly specified");
        return;
     }

     switch(CurrDataStruct) {
       case naive:
         t1=current_time();        
         Found = NaiveNNQuery(Points,numpoints,dimension,querpoint,NumNN,Metric,MinkP);
         t2=current_time();
         if (Timing) {printf("Naive NN query took %f\n",t2-t1);}
         break;
       case optkd:
         if (ShowWhat==DATASTRUCT) {
           if (CurrDisplayMode==twod) {
             Found = DrawkdOptNNQuery2d(Points,numpoints,dimension,querpoint,NumNN,Metric,OptkdRoot, MinkP,XProjDim,YProjDim);
           } else if (CurrDisplayMode==threed) {
             Found = DrawkdOptNNQuery3d(Points,numpoints,dimension,querpoint,NumNN,Metric,OptkdRoot, MinkP,XProjDim,YProjDim,ZProjDim);           
           }
         } else {

           t1=current_time();        
           Found = kdOptNNQuery(Points,dimension,querpoint,NumNN,Metric,OptkdRoot,MinkP);
           t2=current_time();
           if (Timing) {printf("Optkd NN query took %f\n",t2-t1);}
         }
         break;
       case sproullkd:
         t1=current_time();        
         Found = kdSproullNNQuery(Points,dimension,querpoint,NumNN,Metric,SproullkdRoot,MinkP);
         t2=current_time();
         if (Timing) {printf("Sproullkd NN query took %f\n",t2-t1);}
         break;
       case vptree:
         t1=current_time();        
         Found = vpNNQuery(Points,dimension,querpoint,NumNN,Metric,VPRoot);
         t2=current_time();
         if (Timing) {printf("vptree NN query took %f\n",t2-t1);}
         break;
         break;
     }

    if (CurrDisplayMode == twod) {
      DisplayNNQuery2d(Points,Found,querpoint,XProjDim,YProjDim,RED);
    }

    if (CurrDisplayMode == threed) {
      DisplayNNQuery3d(Points,Found,querpoint,XProjDim,YProjDim,ZProjDim,RED);
    }
  }

  if (CurrQueryMethod == nearestneighborgraph) {
    if ((sscanf(XmTextFieldGetString(NNGNumPointsText),"%d",&NumNN) != 1) || (NumNN <1) || 
        (NumNN > numpoints)) {
       DisplayError("Number of nnq points is invalid");
       return;
     }
    switch(CurrDataStruct) {
       case naive:
         break;
       case optkd:
         root=OptkdRoot;
         break;
       case sproullkd:
         root=SproullkdRoot;
         break;
       case vptree:
         root=VPRoot;
         break;
    }
    if (CurrDisplayMode == twod) {
      DrawNNG2d(Points,numpoints,dimension,NumNN,Metric,XProjDim,YProjDim,root,MinkP,CurrDataStruct);
    }
    if (CurrDisplayMode == threed) {
      DrawNNG3d(Points,numpoints,dimension,NumNN,Metric,XProjDim,YProjDim,ZProjDim,MinkP,root,CurrDataStruct);
    }

  }
  free(Found);

}

/*************************************************************************/

void KillExistingDataStructure()

/*************************************************************************/

{
  XtSetSensitive(DisplayQuerButt, FALSE);
  if (CurrDataStruct == naivekd) {
    KillNaiveTree(NaivekdRoot);
    NaivekdRoot=NULL;
  }
  if (CurrDataStruct == optkd) {
    KillOptTree(OptkdRoot);
    OptkdRoot=NULL;
  }
  if (CurrDataStruct == sproullkd) {
    KillSproullTree(SproullkdRoot);
    SproullkdRoot=NULL;
  }
  if (CurrDataStruct == vptree) {
    KillVPTree(VPRoot);
    VPRoot=NULL;
  }
}








